#!/usr/bin/perl

if( $#ARGV != 3 )
{
    print "arguments:  aspath-file initial.txt asrelationship.txt tier1as.txt\n"; exit;
}

$aspath_file  = $ARGV[0];
$asrel_file = $ARGV[1];
$newasrel_file = $ARGV[2];
$tier1as = $ARGV[3];
$valley_file = "$newasrel_file.valley";


$statistic_initial = 0;
$statistic_all = 0;
$statistic_infer = 0;
$statistic_remaining = 0;

##========= read Tier 1 AS =============


$string_tier1as = " ";
open(T1AS, "$tier1as") || die("can not read $tier1as");
while(<T1AS>)
{
    chomp($_);
    @r = split(/\s+/, $_);
    if( $r[0]>0 && $string_tier1as !~ / $r[0] / )
    {
        $string_tier1as = join(" ", $string_tier1as, " $r[0] ");
    }
}
close(T1AS);


## get all AS pairs and set their relationships to unknown

open(INPUT, "$aspath_file") || die("can not read $aspath_file");
while(<INPUT>)
{
    chomp($_); $_ =~ tr/ / /s; $_ =~ s/^ //g; $_ =~ s/ $//g;
    $aspaths{$_} = 1;
    @r = split(/\s+/, $_);
    for($i=0; $i<$#r; $i++)
    {
        last if( $r[$i]==0 || $r[$i+1]==0 );
        if( $r[$i]>0 && $r[$i+1]>0 && $r[$i]!=$r[$i+1])
        {
            if( $pairsSETunknown{$r[$i], $r[$i+1]} !=1 )
            {
                $pairsSETunknown{$r[$i], $r[$i+1]} = 1;
                $pairsSETunknown{$r[$i+1], $r[$i]} = 1;
                $asrel{$r[$i], $r[$i+1]} = "-"; $asrel{$r[$i+1], $r[$i]} = "-";
            }
        }
    }
}
close(INPUT);


#===========read known AS relatonship =======
#$cntPartialASRelTotal = 0;
$cntPartialASRelNotCovered = 0;
open(INPUT,"$asrel_file") || die("can not read $asrel_file");
while(<INPUT>)
{
    chomp($_);
    @r = split(/\s+/, $_);
    if( $r[1]>0 && $r[2]>0 && $r[1]!=$r[2] && $asrel{$r[1], $r[2]} eq "-" )
    {
        $asrel{$r[1], $r[2]} = $r[3];
        if( $r[3] eq "C" )
        {
            $asrel{$r[2], $r[1]} = "P";
        }
        elsif ( $r[3] eq "P" )
        {
            $asrel{$r[2], $r[1]} = "C";
        }
        elsif( $r[3] eq "E" || $r[3] eq "S" )
        {
            $asrel{$r[2], $r[1]} = $r[3];
        }
        elsif( $r[3] ne "-" )
        {
            print "warning: ignore $r[1] $r[2] $r[3]\n";
        }
    }
    else
    {
        $cntPartialASRelNotCovered++;
        # print "warning: ignore $tmp_number initial values  :: $r[1] $r[2] $r[3]\n";
    }
}
close(INPUT);
#print "total number of partial information = $cntPartialASRelTotal\n";
print "number of AS pairs that are not in our data set is $cntPartialASRelNotCovered\n";

#
#  update pairsSET
#
foreach $key ( keys %asrel )
{
    ($as1, $as2) = split($;, $key);
    UpdateSET($as1, $as2, $asrel{$key});
}


#
# get the number of unknown relationship of AS pairs
#
$oldnumSETunknown = 0;
foreach $key ( keys %pairsSETunknown )
{
    $oldnumSETunknown++;
}

$runs = 0;

## InferringRules2();

REPEAT:

    $runs++;
    print "\ndebug...  parsing all AS paths at $runs runs... \n";

    #  put more unknown AS relationship pairs into pairsSET
    print "  ===>  inferring rules...\n";

    InferringRules1();

    #  update AS relationship based on pairsSET
    print "  ===>  refreshing rules...\n";
    RefreshASrel();

    #  update pairsSET
    delete @pairsSETpc{keys %pairsSETpc};
    delete @pairsSETcp{keys %pairsSETcp};
    delete @pairsSETpp{keys %pairsSETpp};
    delete @pairsSETss{keys %pairsSETss};
    delete @pairsSETunknown{keys %pairsSETunknown};
    foreach $key ( keys %asrel )
    {
        ($as1, $as2) = split($;, $key);
        UpdateSET($as1, $as2, $asrel{$key});
    }

    open(OUT, ">$newasrel_file.$runs") || die("can not write $newasrel_file.$runs");
    foreach $key ( keys %asrel )
    {
        ($as1, $as2) = split($;, $key);
        if( $as1<$as2 ) { print OUT " $as1 $as2 $asrel{$key} \n"; }
    }
    close(OUT);

    $newnumSETunknown = 0;
    foreach $key ( keys %pairsSETunknown )
    {
        $newnumSETunknown++;
    }

    print "debug:: old_unknown = $oldnumSETunknown, new_unknown = $newnumSETunknown \n";
    if( $newnumSETunknown < $oldnumSETunknown )
    {
        $oldnumSETunknown = $newnumSETunknown;
        print "  ===>  filtering rules...\n";
        FilteringValleyPath();
        goto REPEAT;
    }

print " Inferring remaining AS relationships\n";

InferringByDegree();

system("cp $newasrel_file.$runs.refined-peer $newasrel_file ");

exit;

sub UpdateSET {
    local($as1, $as2, $rel) = @_;
    if( $as1>0 && $as2>0 && $as1!=$as2 )
    {
        if( $rel eq "P" || $rel eq "C" || $rel eq "E" || $rel eq "S" )
        {
            delete $pairsSETunknown{$as1, $as2};
            delete $pairsSETunknown{$as2, $as1};
        }

        if( $rel eq "P" )
        {
            $pairsSETpc{$as1, $as2} = 1; $pairsSETcp{$as2, $as1} = 1;
            if( $string_tier1as =~ / $as2 / )
            {
                $pairsSETcp{$as1, $as2} = 1; $pairsSETpc{$as2, $as1} = 1;
            }
        }
        elsif( $rel eq "C" )
        {
            $pairsSETcp{$as1, $as2} = 1; $pairsSETpc{$as2, $as1} = 1;
            if( $string_tier1as =~ / $as1 / )
            {
                $pairsSETpc{$as1, $as2} = 1; $pairsSETcp{$as2, $as1} = 1;
            }
        }
        elsif( $rel eq "E" )
        {
            $pairsSETpp{$as1, $as2} = 1; $pairsSETpp{$as2, $as1} = 1;
        }
        elsif( $rel eq "S" )
        {
            $pairsSETss{$as1, $as2} = 1; $pairsSETss{$as2, $as1} = 1;
        }
        else
        {
            $pairsSETunknown{$as1, $as2} = 1; $pairsSETunknown{$as2, $as1} = 1;
        }
    }
    else
    {
       # print "Warning 1111 ::  $as1 $as2 $rel\n";
    }
}


sub RefreshASrel{
    foreach $key (keys %asrel)
    {
        ($as1, $as2) = split($;, $key);
        if( $as1>0 && $as2>0 && $as1 < $as2 )
        {
            if( $pairsSETss{$as1, $as2} == 1 )
            {
                $asrel{$as1, $as2} = "S"; $asrel{$as2, $as1} = "S";
            }
            elsif( $pairsSETpp{$as1, $as2} == 1 )
            {
                $asrel{$as1, $as2} = "E"; $asrel{$as2, $as1} = "E";
            }
            elsif( $pairsSETpc{$as1, $as2} == 1 && $pairsSETcp{$as1, $as2} == 1 )
            {
                $asrel{$as1, $as2} = "S"; $asrel{$as2, $as1} = "S";
            }
            elsif( $pairsSETpc{$as1, $as2} == 1 )
            {
                $asrel{$as1, $as2} = "P"; $asrel{$as2, $as1} = "C";
            }
            elsif( $pairsSETcp{$as1, $as2} == 1 )
            {
                $asrel{$as1, $as2} = "C"; $asrel{$as2, $as1} = "P";
            }
            elsif( $pairsSETunknown{$as1, $as2} == 1 )
            {
                $asrel{$as1, $as2} = "-"; $asrel{$as2, $as1} = "-";
            }
            else
            {
                # print "warning:: in refreshing rules :: $as1 $as2\n";
            }
        }
        elsif( $as1>0 && $as2>0 && $as1>$as2 )
        {
        }
        else
        {
            # print "warning 2222 :: as1 = $as1 as2 = $as2 asrel = $asrel{$as1, $as2}\n";
        }
    }
}

sub FilteringValleyPath
{
    local($total, $count_filter, $count_passed);

    $total = 0; $count_filter = 0; $count_passed = 0;

    open(VALLEY, ">$valley_file.$runs") || die("can not write $valley_file.$runs");

    foreach $key ( keys %aspaths )
    {
        $total++;

        $valley = " ";
        @r = split(/\s+/, $key);

        $toptier = " ";
        $min = $#r + 1; $max = -1;
        for($i=0; $i<=$#r; $i++)
        {
            last if ($r[$i]==0);
            if( $r[$i]>0 && $string_tier1as =~ / $r[$i] / )
            {
                $toptier = join(" ", $toptier, "T");
                if( $max<=$i )
                {
                    $max = $i;
                }
                if( $min>=$i )
                {
                    $min = $i;
                }
            }
            elsif ( $r[$i]>0 )
            {
                $toptier = join(" ", $toptier, "o");
            }
        }

        $toptier =~ s/ //g;
        $interval = $max - $min;

        for($i=0; $i<$#r; $i++)
        {
            last if( $r[$i]==0 || $r[$i+1]==0 );
            if( $r[$i]>0 && $r[$i+1]>0 && $r[$i]!=$r[$i+1])
            {
                $valley = join(" ", $valley, " $asrel{$r[$i], $r[$i+1]} ");
            }
        }

        $ori_valley = $valley;
        $ori_valley =~ tr/ / /s;
        $valley =~ s/-/ /g;
        $valley =~ s/S/ /g;
        $valley =~ s/ //g;
        $valley =~ tr/C/C/s;
        $valley =~ tr/P/P/s;

        if( $valley eq "C" || $valley eq "CP" || $valley eq "CE"  || $valley eq "CEP"
            ||$valley eq "E" || $valley eq "EP" || $valley eq "P" || $valley eq "")
        {
            $count_passed;
        }
        else
        {
            print VALLEY " ::valley:: $ori_valley :: $toptier ::   $key  \n";
            delete $aspaths{$key};
            $count_filter++;
        }
    }
    print VALLEY " ::summary:: total $total  passed $count_passed  filtered: $count_filter \n";
    close(VALLEY);
}

#
#   valley free applied by initial value
#
sub InferringRules1
{
    foreach $key ( keys %aspaths )
    {
        $lastCP = -1; $firstPP = -1; $lastPP = -1; $firstPC = -1;
        $firstTop = -1; $lastTop = -1;

        @r = split(/\s+/, $key);
        for($i=0; $i<=$#r; $i++)
        {
            last if( $r[$i]==0 );
            if( $string_tier1as =~ / $r[$i] / )
            {
                if( $firstTop ==-1 ) { $firstTop = $i; $lastTop = $i;}
                else { $lastTop = $i; }
            }
        }

        @r = split(/\s+/, $key);
        for($i=0; $i<$#r; $i++)
        {
            last if( $r[$i]==0 || $r[$i+1]==0 );
            if( $r[$i]>0 && $r[$i+1]>0 && $r[$i]!=$r[$i+1])
            {
                if( $asrel{$r[$i], $r[$i+1]} eq "C" ) { $lastCP = $i; }
                if( $firstPC ==-1 && $asrel{$r[$i], $r[$i+1]} eq "P" ) { $firstPC = $i; }
                if( $asrel{$r[$i], $r[$i+1]} eq "E" )
                {
                    if( $firstPP ==-1 ) { $firstPP = $i; $lastPP = $i;}
                    else
                    {
                      $lastPP = $i;
                      # print "warning $firstPP $lastPP peers :: $key\n";
                    }
                }
            }
        }

        $valley = 0;

        if( $firstTop >= 0 && $lastCP > $firstTop )
        {
            # print "warning:: valley path :: Top $firstTop CP $lastCP :: $key\n";
            $valley = 1;
        }

        if( $firstPC >= 0 && $firstPC < $lastTop )
        {
            # print "warning:: valley path :: Top $lastTop PC $firstPC :: $key\n";
            $valley = 1;
        }

        if( $firstPC >= 0 )
        {
            if( $lastCP > $firstPC || $lastPP > $firstPC )
            {
                # print "warning:: valley path :: PC $firstPC CP $lastCP PP $lastPP :: $key\n";
                $valley = 1;
            }
        }

        if( $firstPP >= 0 )
        {
            if( $lastCP > $firstPP )
            {
                # print "warning:: valley path :: PC $firstPC CP $lastCP PP $firstPP :: $key\n";
                $valley = 1;
            }
        }

        ##   E/P can only followed by P/S
        #
        if( $valley ==0 )
        {
            ## all asrelationship ahead CP is CP
            for($i=0; $i<$lastCP && $lastCP>=0; $i++)
            {
                last if( $r[$i]==0 || $r[$i+1]==0 );
                if( $r[$i] != $r[$i+1] )
                {
                    UpdateSET($r[$i], $r[$i+1], "C");
                }
            }

            ## all asrelationship behind PC is PC
            for($i=$firstPC+1; $i<$#r && $firstPC>=0; $i++)
            {
                last if( $r[$i]==0 || $r[$i+1]==0 );
                if( $r[$i] != $r[$i+1] )
                {
                    UpdateSET($r[$i], $r[$i+1], "P");
                }
            }

            ## all asrelationship ahead PP is CP
            for($i=0; $i<$firstPP && $firstPP>=0; $i++)
            {
                last if( $r[$i]==0 || $r[$i+1]==0 );
                if( $r[$i] != $r[$i+1] )
                {
                    UpdateSET($r[$i], $r[$i+1], "C");
                }
            }

            ## all asrelationship behind PP is PC
            for($i=$lastPP+1; $i<$#r && $lastPP>=0; $i++)
            {
                last if( $r[$i]==0 || $r[$i+1]==0 );
                if( $r[$i] != $r[$i+1] )
                {
                    UpdateSET($r[$i], $r[$i+1], "P");
                }
            }
        }
    }
}


#
#   valley free applied by top provider
#   TopProviders are in "string_tier1as"
#
sub InferringRules2
{
    foreach $key ( keys %aspaths )
    {
        $firstTop = -1; $lastTop = -1;
        @r = split(/\s+/, $key);
        for($i=0; $i<=$#r; $i++)
        {
            last if( $r[$i]==0 );
            if( $string_tier1as =~ / $r[$i] / )
            {
                if( $firstTop ==-1 ) { $firstTop = $i; $lastTop = $i;}
                else { $lastTop = $i; }
            }
        }

        ##   E/P can only followed by P/S
        #
        if( $firstTop >= 0 )
        {
            ## all asrelationship ahead first Top is CP
            for($i=0; $i<$firstTop; $i++)
            {
                last if( $r[$i]==0 || $r[$i+1]==0 );
                if( $r[$i] != $r[$i+1] )
                {
                    UpdateSET($r[$i], $r[$i+1], "C");
                }
            }
        }

        if( $lastTop>=0 )
        {
            ## all asrelationship behind last Top is PC
            for($i=$lastTop; $i<$#r; $i++)
            {
                last if( $r[$i]==0 || $r[$i+1]==0 );
                if( $r[$i] != $r[$i+1] )
                {
                    UpdateSET($r[$i], $r[$i+1], "P");
                }
            }
        }
    }
}

sub InferringByDegree
{
    #
    #  get degree for each AS
    #

    foreach $key ( keys %asrel )
    {
        ($as1, $as2) = split($;, $key);
        if( $as1<$as2 )
        {
            $degree{$as1}++;
            $degree{$as2}++;
        }
    }

    foreach $key ( keys %aspaths )
    {
        # get left and right index, and find the index of maximum degree

        # 0o1o2o3-4-5-6-7o8o9  then left=3 right=7
        $left = -1; $right = -1;
        @r = split(/\s+/, $key);
        for($i=0; $i<$#r; $i++)
        {
            last if( $r[$i]==0 || $r[$i+1]==0 );
            if( $r[$i]>0 && $r[$i+1]>0 && $asrel{$r[$i], $r[$i+1]} eq "-" )
            {
                $left = $i;
                last;
            }
        }

        for($i=0; $i<$#r; $i++)
        {
            last if( $r[$i]==0 || $r[$i+1]==0 );
            if( $r[$i]>0 && $r[$i+1]>0 && $asrel{$r[$i], $r[$i+1]} eq "-" )
            {
                $right = $i;
            }
        }
        $right++;

        $left{$key} = $left; $right{$key} = $right;

        ##  if some unknown AS relationship pairs exist in this path
        if( $left>=0 && $right>=0 )
        {
            $max = $left;
            for($i=$left; $i<=$right; $i++)
            {
                if( $degree{$r[$i]}>$degree{$r[$max]} )
                {
                    $max = $i;
                }
            }
            # inferring transit flow direction
            for($i=$left; $i<$max; $i++)
            {
                $transit{$r[$i], $r[$i+1]}++;
                $remaining{$r[$i], $r[$i+1]} = 1;
                $remaining{$r[$i+1], $r[$i]} = 1;
            }
            for($i=$right; $i>$max; $i--)
            {
                $transit{$r[$i], $r[$i-1]}++;
                $remaining{$r[$i], $r[$i-1]} = 1;
                $remaining{$r[$i-1], $r[$i]} = 1;
            }
        }
    }

    ## infering Provider/Customer and Sibling
    $L = 1;
    foreach $key (  keys %transit )
    {
        ($v1, $v2) = split( $;, $key );
        if( ( $transit{$v1, $v2}>$L && $transit{$v2, $v1}>$L )
         || ( $transit{$v1, $v2}<=$L && $transit{$v1, $v2}>0
         && $transit{$v2, $v1}<=$L && $transit{$v2, $v1}>0 )  # updated on 02/27/2006
         )
        {
            $asrel{$v1, $v2} = "S";
            $asrel{$v2, $v1} = "S";
        }
        elsif ( $transit{$v1, $v2}>$L || $transit{$v2, $v1}==0 )
        {
            $asrel{$v1, $v2} = "C";
            $asrel{$v2, $v1} = "P";
            if( $string_tier1as =~ / $v1 / )
            {
                ## if two ASes are in Tier 1, mark the link as peer
                ## otherwise, mark the lik as sibling
                # updated on 02/27/2006
                if( $string_tier1as =~ / $v2 / )
                {
                    $asrel{$v1, $v2} = "E";
                    $asrel{$v2, $v1} = "E";
                }
                else
                {
                    $asrel{$v1, $v2} = "S";
                    $asrel{$v2, $v1} = "S";
                }
            }
        }
        elsif ($transit{$v2, $v1}>$L || $transit{$v1, $v2}==0 )
        {
            $asrel{$v1, $v2} = "P";
            $asrel{$v2, $v1} = "C";
            if( $string_tier1as =~ / $v2 / )
            {
                ## if two ASes are in Tier 1, mark the link as peer
                ## otherwise, mark the lik as sibling
                # updated on 02/27/2006
                if( $string_tier1as =~ / $v1 / )
                {
                    $asrel{$v1, $v2} = "E";
                    $asrel{$v2, $v1} = "E";
                }
                else
                {
                    $asrel{$v1, $v2} = "S";
                    $asrel{$v2, $v1} = "S";
                }
            }
        }
        else
        {
            print "\nDEBUG other unknown case \n";
        }
    }

    open(OUT, ">$newasrel_file.$runs.refined-sibling") || die("can not write $newasrel_file.$runs.refined-sibling");
    foreach $key ( keys %asrel )
    {
        ($as1, $as2) = split($;, $key);
        if( $as1<$as2 ) { print OUT " $as1 $as2 $asrel{$key} \n"; }
    }
    close(OUT);


    #  Inferring Peer Heuristic Algorithm in Lixin Gao's Paper

    # first, finding not peering
    foreach $key ( keys %aspaths )
    {
        if( $left{$key} >= 0 && $right{$key} >=0 )
        {
            $left = $left{$key}; $right = $right{$key};

            # find the index of top AS in AS path
            $top = -1;

            if( $asrel{$r[$left], $r[$left+1]} eq "P" )
            {
                $top = $left;
            }
            elsif( $asrel{$r[$right-1], $r[$right]} eq "C" )
            {
                $top = $right;
            }
            else
            {
                for($i=$left; $i<$right-1; $i++)
                {
                    if( $asrel{$r[$i], $r[$i+1]} eq "C" && $asrel{$r[$i+1], $r[$i+2]} eq "P" )
                    {
                        $top = $i+1;
                    }
                }
            }

            if( $top==-1 )
            {
                for($i=$left; $i<$right-1; $i++)
                {
                    $notpeering{$r[$i], $r[$i+1]} = 1;
                }
            }
            else
            {
                for($i=$left; $i<$top-1; $i++)
                {
                    $notpeering{$r[$i], $r[$i+1]} = 1;
                }
                for($i=$top+1; $i<$right-1; $i++)
                {
                    $notpeering{$r[$i], $r[$i+1]} = 1;
                }
                ## mark the link with lower degree as nonpeering
                # updated on 02/27/2006
                if( $degree{$r[$top-1]} > $degree{$r[$top+1]} )
                {
                    $notpeering{$r[$top], $r[$top+1]} = 1;
                }
                else
                {
                    $notpeering{$r[$top-1], $r[$top]} = 1;
                }
            }
        }
    }

    # second, inferring peering relationship
    $R = 60;
    foreach $key (  keys %remaining )
    {
        ($v1, $v2) = split( $;, $key );
        # print " debug $v1 $v2 degree $degree{$v1}, $degree{$v2} ";
        if( $degree{$v1} >= $degree{$v2} ) { $d1 = $degree{$v1}; $d2 = $degree{$v2}; }
        else { $d1 = $degree{$v2}; $d2 = $degree{$v1}; }
        $ratio = $d1/$d2;

        if( $notpeering{$v1, $v2}!=1 && $notpeering{$v2, $v1}!=1
            && $degree{$v1}/$degree{$v2} < $R && $degree{$v1}/$degree{$v2} > 1/$R )
        {
            # print " assigned($asrel{$key}) to E ($ratio)\n";
            $asrel{$v1, $v2} = "E";
            $asrel{$v2, $v1} = "E";
        }
        else
        {
           # print " remaining to $asrel{$v1, $v2} ($ratio)\n";
        }
    }

    open(OUT, ">$newasrel_file.$runs.refined-peer") || die("can not write $newasrel_file.$runs.refined-peer");
    foreach $key ( keys %asrel )
    {
        ($as1, $as2) = split($;, $key);
        if( $as1<$as2 ) { print OUT " $as1 $as2 $asrel{$key} \n"; }
    }
    close(OUT);
}


sub numerically  {         $a  <=>         $b ;}
sub cnumerically { $asccnt{$b} <=> $asccnt{$a};}
